@@ -5,8 +5,6 @@ This module provides access to a virtual data lake via its web API.
 import json
 import requests
 from . import parts
-api = None
-session = requests.session()
 
 
 def public_level():
@@ -17,7 +15,7 @@ def public_level():
       * an `Itm` object representing the public access level.
     '''
     
-    return Itm('0:1')
+    return Itm('0_1')
 
 
 def source_admin_level(source):
@@ -34,140 +32,196 @@ def source_admin_level(source):
     return Itm(repr(source) + '_-9223372036854775807')
 
 
-def set_api(url):
-    '''Set the URL for access to the API.
-    
-    Parameters:
-      * url(str): the URL of the system's API.
+class Client:
     '''
-    global api
-    api = url
+    An API client that can be used to invoke API methods.
+    '''
+    
+    def __init__(self, url):
+        '''
+        Construct an API client. The client will have a session
+        that is not authenticated, so that API methods that
+        require authentication must have credential_and_key
+        parameters to provide it. (The `set_session` function
+        can be used to replace the session by a session that has 
+        been authenticated.)
+        
+        Parameters:
+          * url (String) the API's URL.
+        '''
+        
+        self.api = url
+        self.session = requests.session()
 
 
-def set_session(session_to_be_set):
-    '''Set the session for access to the API.
+    def set_session(self, session_to_be_set):
+        '''Set the session for access to the API.
     
-    This method enables a session that has been authenticated
-    by means outside this module to be used for API interactions.
+        This method enables a session that has been authenticated
+        by means outside this module to be used for API interactions.
     
-    Parameters:
-      * session_to_be_set: the requests session to be used for
-      interactions with the system's API.
-    '''
-    global session
-    session = session_to_be_set
+        Parameters:
+          * session_to_be_set: the requests session to be used for
+          interactions with the system's API.
+        '''
+        self.session = session_to_be_set
     
     
-def query(constraints, credential_and_key=None):
-    '''
-    Get data that satisfies a set of constraints
+    def query(self, constraints, credential_and_key=None):
+        '''
+        Get data that satisfies a set of constraints
     
-    Parameters:
-      * constraints: a list of constraints that the data must satisfy.
-        Each constraint can be supplied either as a `Constraint` object
-        or as a 3-tuple whose elements are the subject, verb and object
-        of a constraint. 
-      * credential_and_key: a 2-tuple whose first element is an `Itm`
-        object that is a credential to authorize the request, and
-        whose second element is the credential's key.
-        If a credential_and_key is not supplied, the request will be 
-        made at the lowest (public) level.
+        Parameters:
+          * constraints: a list of constraints that the data must satisfy.
+            Each constraint can be supplied either as a `Constraint` object
+            or as a 3-tuple whose elements are the subject, verb and object
+            of a constraint. 
+          * credential_and_key: a 2-tuple whose first element is either an
+            `Itm` object that is a credential to authorize the request or
+            the identifier of such an object, and whose second element is 
+            the credential's key. If a credential_and_key is not supplied, 
+            and the `set_session` function has been invoked to set a
+            pre-authenticated session, the request will be made at its 
+            access level. If a credential_and_key is not supplied, 
+            and no pre-authenticated session has been set, the request
+            will be made at the lowest (public) level.
       
-    Returns:
-      * a list of dictionaries, each of which contains an assignment
-      of values to unknowns that satisfies the constraints.
-    '''
-    
-    parms = {'command': 'query'}
-    scanned_constraints = []
-    files = dict()
-    for input_constraint in constraints:
-        constraint = __get_constraint(input_constraint)
-        scanned_constraints.append(
-            [
-                __get_quantum(constraint.subj, files),
-                __get_quantum(constraint.vrb, files),
-                __get_quantum(constraint.obj, files)
-            ])
-        
-    files.update({"constraints": ("constraints", json.dumps(scanned_constraints), "text/plain; charset=utf-8")})
-    if credential_and_key is not None:
-        auth = {'credential': credential_and_key[0].itemid, 'key': credential_and_key[1]}
-        files.update({"auth": ("auth", json.dumps(auth), "text/plain; charset=utf-8")})
-    response = session.post(api, files=files, params=parms)
-    if response.status_code == 200: return __get_assignments_dict_list(response)
-    else: __handle_http_error(response)
-    
-    
-def update(changes, credential_and_key=None):
-    '''
-    Make a set of virtual data lake changes.
-        
-    Parameters:
-      * changes: a `Change` object or a list of `Change` objects
-      * credential_and_key: a 2-tuple whose first element is an `Itm`
-        object that is a credential to authorize the request, and
-        whose second element is the credential's key.
-        If a credential_and_key is not supplied, the request will be 
-        made at the lowest (public) level.
+        Raises:
+          * VdlException: if the query cannot be executed.
       
-    Raises:
-      * VdlException: if one of the commands cannot be executed,
-        or if an error occurred in transmitting the request or in
-        receiving or decoding the response. Depending on the nature
-        of the error, either all of the commands or none of the
-        commands will have been executed.
+        Returns:
+          * a list of dictionaries, each of which contains an assignment
+          of values to unknowns that satisfies the constraints.
+        '''
+    
+        parms = {'command': 'query'}
+        scanned_constraints = []
+        files = dict()
         
-    Returns:
-      * The assignments of items to the handles of the commands, 
-        as a dictionary.  
+        for input_constraint in constraints:
+            constraint = __get_constraint__(input_constraint)
+            scanned_constraints.append(
+                [
+                    __get_quantum__(constraint.subj, files),
+                    __get_quantum__(constraint.vrb, files),
+                    __get_quantum__(constraint.obj, files)
+                    ])
+        
+        files.update({"constraints": ("constraints", json.dumps(scanned_constraints), "text/plain; charset=utf-8")})
+        
+        if credential_and_key is not None:
+            auth = __get_credential_string_and_key__(credential_and_key)
+            files.update({"auth": ("auth", json.dumps(auth), "text/plain; charset=utf-8")})
+            
+        response = self.session.post(self.api, files=files, params=parms)
+        if response.status_code == 200: return __get_assignments_dict_list__(response)
+        else: __handle_http_error__(response)
+    
+    
+    def update(self, changes, credential_and_key=None):
+        '''
+        Make a set of virtual data lake changes.
+        
+        Parameters:
+          * changes: a `Change` object or a list of `Change` objects
+          * credential_and_key: a 2-tuple whose first element is either an
+            `Itm` object that is a credential to authorize the request or
+            the identifier of such an object, and whose second element is 
+            the credential's key. If a credential_and_key is not supplied, 
+            and the `set_session` function has been invoked to set a
+            pre-authenticated session, the request will be made at its 
+            access level. If a credential_and_key is not supplied, 
+            and no pre-authenticated session has been set, the request
+            will be made at the lowest (public) level (and will fail).
+      
+        Raises:
+          * VdlException: if one of the commands cannot be executed,
+            or if an error occurred in transmitting the request or in
+            receiving or decoding the response. Depending on the nature
+            of the error, either all of the commands or none of the
+            commands will have been executed.
+        
+        Returns:
+          * The assignments of items to the handles of the commands, 
+            as a dictionary.  
+        '''
+    
+        if isinstance(changes, Change): changes = [changes]
+        parms = {'command': 'update'}
+        scanned_changes = []
+        files = dict()
+        for change in changes:
+            scanned_change = {}
+        
+            if isinstance(change, CreateItem):
+                scanned_change['change'] = 'create item'
+                scanned_change['handle'] = change.handle
+                scanned_change['source'] = change.source
+                if change.read is not None: scanned_change['read'] = change.read.itemid
+                if change.write is not None: scanned_change['write'] = change.write.itemid
+        
+            elif isinstance(change, UpdateItem):
+                scanned_change['change'] = 'update item'
+                scanned_change['item'] = change.item.itemid
+                if change.read is not None: scanned_change['read'] = change.read.itemid
+                if change.write is not None: scanned_change['write'] = change.write.itemid
+        
+            elif isinstance(change, DeleteItem):
+                scanned_change['change'] = 'delete item'
+                scanned_change['item'] = change.item.itemid
+        
+            elif isinstance(change, PutTriple):
+                scanned_change['change'] = 'put triple'
+                scanned_change['subject'] = __get_quantum__(change.subj, files)
+                scanned_change['verb'] = __get_quantum__(change.vrb, files)
+                scanned_change['object'] = __get_quantum__(change.obj, files)
+        
+            elif isinstance(change, RemoveTriples):
+                scanned_change['change'] = 'remove triples'
+                scanned_change['subject'] = repr(change.subj)
+                scanned_change['verb'] = repr(change.vrb)
+                if change.obj is not None: scanned_change['object'] = __get_datum__(change.obj, files)
+        
+            elif isinstance(change, SetUniqueObject):
+                scanned_change['change'] = 'set unique object'
+                scanned_change['subject'] = repr(change.subj)
+                scanned_change['verb'] = repr(change.vrb)
+                scanned_change['object'] = __get_quantum__(change.obj, files)
+            
+            scanned_changes.append(scanned_change)
+            
+        files.update({"changes": ("changes", json.dumps(scanned_changes), "text/plain; charset=utf-8")})
+        
+        if credential_and_key is not None:
+            auth = __get_credential_string_and_key__(credential_and_key)
+            files.update({"auth": ("auth", json.dumps(auth), "text/plain; charset=utf-8")})
+            
+        response = self.session.post(self.api, files=files, params=parms)
+        if response.status_code == 200: return __get_items_dict__(response)
+        else: __handle_http_error__(response)
+
+
+def __get_credential_string_and_key__(credential_and_key):
+    '''
+    Get a credential and key as dictionary entries.
+    
+    Parameters:
+      * credential_and_key: a tuple whose first element is an item
+        representing a credential, and whose second element is a string 
+        that is a credential key
+        
+    Raises: VdlException if a tuple element is not valid
+    
+    Returns: a dictionary with entry 'credential' containing the identifier
+        of the item representing the credential, and 'key' containing the
+        key.
     '''
     
-    if isinstance(changes, Change): changes = [changes]
-    parms = {'command': 'update'}
-    scanned_changes = []
-    files = dict()
-    for change in changes:
-        scanned_change = {}
-        
-        if isinstance(change, CreateItem):
-            scanned_change['change'] = 'create item'
-            scanned_change['handle'] = change.handle
-            scanned_change['source'] = change.source
-            if change.read is not None: scanned_change['read'] = change.read.itemid
-            if change.write is not None: scanned_change['write'] = change.write.itemid
-        
-        elif isinstance(change, DeleteItem):
-            scanned_change['change'] = 'delete item'
-            scanned_change['item'] = change.item.itemid
-        
-        elif isinstance(change, PutTriple):
-            scanned_change['change'] = 'put triple'
-            scanned_change['subject'] = __get_quantum(change.subj, files)
-            scanned_change['verb'] = __get_quantum(change.vrb, files)
-            scanned_change['object'] = __get_quantum(change.obj, files)
-        
-        elif isinstance(change, RemoveTriples):
-            scanned_change['change'] = 'remove triples'
-            scanned_change['subject'] = repr(change.subj)
-            scanned_change['verb'] = repr(change.vrb)
-            if change.obj is not None: scanned_change['object'] = __get_datum(change.obj, files)
-        
-        elif isinstance(change, SetUniqueObject):
-            scanned_change['change'] = 'set unique object'
-            scanned_change['subject'] = repr(change.subj)
-            scanned_change['verb'] = repr(change.vrb)
-            scanned_change['object'] = __get_quantum(change.obj, files)
-            
-        scanned_changes.append(scanned_change)
-            
-    files.update({"changes": ("changes", json.dumps(scanned_changes), "text/plain; charset=utf-8")})
-    if credential_and_key is not None:
-        auth = {'credential': credential_and_key[0].itemid, 'key': credential_and_key[1]}
-        files.update({"auth": ("auth", json.dumps(auth), "text/plain; charset=utf-8")})
-    response = session.post(api, files=files, params=parms)
-    if response.status_code == 200: return __get_items_dict(response)
-    else: __handle_http_error(response)
+    cred = credential_and_key[0]
+    if isinstance(cred, Itm): cred_string = cred.itemid
+    elif isinstance(cred, str): cred_string = cred
+    else: raise VdlException("Invalid credential: " + str(cred));           
+    return {'credential': cred_string, 'key': credential_and_key[1]}
 
 
 class Constraint:
@@ -239,7 +293,7 @@ class Constraint:
         return "CONSTRAINT(" + self.subj.__repr__() + ', ' + self.vrb.__repr__() + ', ' + self.obj.__repr__() + ')'
         
         
-def __get_constraint(constraint_or_tuple):
+def __get_constraint__(constraint_or_tuple):
     '''
     Get a constraint that is supplied either as a 'Constraint' object
     or as a 3-tuple whose elements are the constraint's subject, verb
@@ -262,7 +316,7 @@ def __get_constraint(constraint_or_tuple):
     if isinstance(constraint_or_tuple, tuple) and len(constraint_or_tuple) == 3: 
         return Constraint(constraint_or_tuple[0], constraint_or_tuple[1], constraint_or_tuple[2])
     raise VdlException("Not a constraint: " + str(constraint_or_tuple))
-    
+      
 
 class Change:
     '''
@@ -327,6 +381,51 @@ class CreateItem(Change):
         return "CREATE ITEM " + self.handle + ' IN SOURCE ' + str(self.source) + prread + prwrite
     
     
+class UpdateItem(Change):
+    '''
+    A virtual data lake change that updates an item, changing its access levels.
+    '''
+    
+    def __init__(self, item, read, write=None):
+        '''
+        Construct a virtual data lake change that updates an item.
+        
+        Parameters:
+          * item(`Itm`) the item to be updated.
+          * read: the read access level that the item is to have. 
+            This can be an `Itm` object or an `ItmName` object. 
+            If it is None then the item's read level is not changed.
+          * write: the write access level that the item is to have. 
+            This can be an `Itm` object or an `ItmName` object. 
+            If it is not supplied or None then the item's write level
+            is not changed.
+            
+        Raises: 
+          * VdlException if the item is not an `Itm` object, or if the
+            read and write levels are both None.
+        '''
+        
+        if not isinstance(item, Itm): raise VdlException('Not a valid item ' + str(item))
+        if read is None and write is None: 
+            raise VdlException('A new read level or a new write level must be supplied')
+        self.item = item
+        self.read = read
+        self.write = write
+        
+        
+    def __repr__(self):
+        if self.read is None: prread = None
+        else: prread = "READ LEVEL " + str(self.read)
+        if self.write is None: prwrite = None
+        else: prwrite = "WRITE LEVEL " + str(self.write)
+        ret = "UPDATE ITEM " + self.item + ' WITH ' 
+        if prread is None: ret = ret + prwrite
+        else:
+            ret = ret + prread
+            if prwrite is not None: ret = ret + ' AND ' + prwrite
+        return ret
+    
+    
 class DeleteItem(Change):
     '''
     A virtual data lake change that deletes an item.
@@ -362,18 +461,18 @@ class PutTriple(Change):
         
         Parameters:
           * subj: the subject. This can be specified by an `Itm` object, 
-            an `ItmName` object, or a handle of an object created in a 
-            previous change.
+            an `ItmName` object, or an `Unknown` that is the handle of 
+            an object created in a previous change.
           * vrb: the verb. This can be specified by an `Itm` object, 
-            an `ItmName` object, or a handle of an object created in a 
-            previous change.
+            an `ItmName` object, or an `Unknown` that is the handle of 
+            an object created in a previous change.
           * obj: the object. This can be an item, a Boolean, an integer, 
             a real number, a text object or a binary object. An item is
-            specified by an `Itm` object, an `ItmName` object, or a 
-            handle of an object created in a previous change. A Boolean
-            is given as a bool, an integer as an int, a real number as a
-            float, a text object as a string, and a binary object as either
-            a bytes or a bytearray.
+            specified by an `Itm` object, an `ItmName` object, or an
+            `Unknown` that is the handle of an object created in a 
+            previous change. A Boolean is given as a bool, an integer as 
+            an int, a real number as a float, a text object as a string, 
+            and a binary object as either a bytes or a bytearray.
         '''
     
         self.subj = subj
@@ -584,7 +683,7 @@ class VdlException(RuntimeError):
         super().__init__(*messages)
        
   
-def __get_assignments_dict(response):
+def __get_assignments_dict__(response):
     '''
     Get a dict that contains a set of assignments of values to 
     unknowns from the content of a request response.
@@ -601,15 +700,15 @@ def __get_assignments_dict(response):
       * a dict containing the set of values.
     '''
     
-    parts = __get_parts(response)
+    parts = __get_parts__(response)
     json_part = parts['json']
     if json_part is None: raise VdlException('No JSON part')
     assignments = json.loads(json_part.content().decode())
     if not isinstance(assignments, dict): raise VdlException('Not a value set')
-    return __get_assigned_values(assignments, parts)
+    return __get_assigned_values__(assignments, parts)
        
   
-def __get_assignments_dict_list(response):
+def __get_assignments_dict_list__(response):
     '''
     Get a list of dicts, each of which contains a set of assignments
     of values to unknowns, from the content of a request response.
@@ -627,7 +726,7 @@ def __get_assignments_dict_list(response):
         of values to unknowns.
     '''
     
-    parts = __get_parts(response)
+    parts = __get_parts__(response)
     json_part = parts['json']
     if json_part is None: raise VdlException('No JSON part')
     assignments_list = json.loads(json_part.content().decode())
@@ -636,11 +735,11 @@ def __get_assignments_dict_list(response):
     values_list = []
     for assignments in assignments_list:
         if not isinstance(assignments, dict): raise VdlException('Not a value set')
-        values_list.append(__get_assigned_values(assignments, parts))
+        values_list.append(__get_assigned_values__(assignments, parts))
     return values_list
     
     
-def __get_assigned_values(assignments_dict, parts):
+def __get_assigned_values__(assignments_dict, parts):
     '''
     Get the assigned values from an assignments dictionary in
     a multipart response.
@@ -710,7 +809,7 @@ def __get_assigned_values(assignments_dict, parts):
     return assigned_values
        
   
-def __get_items_dict(response):
+def __get_items_dict__(response):
     '''
     Get a dictionary keyed by item handles whose values are items
     from the content of a request response.
@@ -728,7 +827,7 @@ def __get_items_dict(response):
       identifiers.
     '''
     
-    parts = __get_parts(response)
+    parts = __get_parts__(response)
     json_part = parts['json']
     if json_part is None: raise VdlException('No JSON part')
     itemids_dict = json.loads(json_part.content().decode())
@@ -739,7 +838,7 @@ def __get_items_dict(response):
     return items_dict
 
 
-def __get_parts(response):
+def __get_parts__(response):
     '''
     Get the parts of a multi-part response
     
@@ -756,15 +855,16 @@ def __get_parts(response):
         parts.part objects
     '''
     
-    content_type = response.headers['content-type']
-    if content_type is None: raise VdlException('No content type header')
+    content_type = response.headers.get('content-type')
+    if content_type is None: raise VdlException(
+        'No content type header. The operation may succeed if re-tried.')
     content_type_line = 'Content-Type: ' + content_type 
     content_type_header = parts.Header(content_type_line)
     boundary = content_type_header.get_parm('boundary')
     return parts.get_parts(response.content, boundary)
        
        
-def __get_quantum(spec, files):
+def __get_quantum__(spec, files):
     '''
     Get the representation of a quantity that can be given as a 
     particular item or piece of data or can be unknown.
@@ -800,10 +900,10 @@ def __get_quantum(spec, files):
     '''
     
     if isinstance(spec, Unknown): return {'unknown': repr(spec)}
-    else: return __get_datum(spec, files)
+    else: return __get_datum__(spec, files)
        
        
-def __get_datum(spec, files):
+def __get_datum__(spec, files):
     '''
     Get the representation of a quantity that is a particular item or 
     piece of data.
@@ -847,7 +947,7 @@ def __get_datum(spec, files):
     else: raise VdlException('Invalid specification: ' + str(spec))
         
                         
-def __handle_http_error(response):
+def __handle_http_error__(response):
     '''
     Handle an error that occurs during an HTTP communication
     with a virtual data lake. The error may occur in transmission
